import os

import uvicorn
from starlette.applications import Starlette


_DEFAULTS = {"port": "8080"}


def run(app: Starlette, **kwargs):
    # The log_config argument takes a logging.dictConfig that is applied to the
    # uvicorn logging.

    overrides = {}
    port = os.environ.get("SERVER_PORT")
    if port is not None:
        overrides["port"] = port
    uvicorn.run(app, **{**_DEFAULTS, **overrides, **kwargs})



import json
from abc import ABC, abstractmethod
from enum import Enum, auto
from itertools import chain
from typing import Callable, Iterator, List, Optional, Tuple

from fastapi.params import Depends
from fastapi.routing import APIRoute
from starlette.requests import Request
from starlette.responses import JSONResponse, Response

from apish.base import Problem
from apish.utils import get_return_type


class CustomRoute(APIRoute):
    MAX_BODY_LENGTH = 400

    def get_route_handler(self) -> Callable:
        original_route_handler = super().get_route_handler()

        async def custom_route_handler(request: Request) -> Response:
            try:
                return await original_route_handler(request)
            except Exception as exc:
                body = await request.body()
                try:
                    request.state.body = json.dumps(body)[: self.MAX_BODY_LENGTH]
                except Exception:  # pylint: disable=broad-except
                    request.state.body = str(body)[: self.MAX_BODY_LENGTH]
                raise exc

        return custom_route_handler


class Verb(Enum):
    GET = auto()
    PUT = auto()
    POST = auto()
    PATCH = auto()
    DELETE = auto()

    @classmethod
    def handler_names(cls) -> Iterator[str]:
        for verb in cls:
            yield verb.handler_name

    @property
    def handler_name(self):
        # pylint: disable=no-member
        return self.name.lower()


class RouteGenerator(ABC):
    @abstractmethod
    def routes(self, tags=None, prefix=None, dependencies=None) -> Iterator[APIRoute]:
        pass


class Namespace(RouteGenerator):
    _children: List[RouteGenerator]

    def __init__(
        self, tags=None, prefix: Optional[str] = None, dependencies: List[Depends] = None
    ):
        self._tags = tags
        self._prefix = prefix
        self._dependencies = dependencies
        self._children = []

    def add(self, route_generator: RouteGenerator) -> None:
        self._children.append(route_generator)

    def route(self, *args, **kwargs):
        def decorator(cls):
            self.add(cls(*args, **kwargs))
            return cls

        return decorator

    def routes(self, tags=None, prefix=None, dependencies=None) -> Iterator[APIRoute]:
        tags = (self._tags or []) + (tags or [])
        prefix = "".join(s for s in [prefix, self._prefix] if s is not None)
        dependencies = (self._dependencies or []) + (dependencies or [])
        return chain(*[child.routes(tags, prefix, dependencies) for child in self._children])


class Resource(RouteGenerator):

    _RESPONSES = {"4XX": {"model": Problem}}

    def __init__(
        self,
        prefix: Optional[str] = None,
        tags: Optional[List[str]] = None,
        dependencies: List[Depends] = None,
    ):
        self._prefix = prefix
        self._tags = tags
        self._dependencies = dependencies

    def handlers(self) -> Iterator[Tuple[Verb, Callable]]:
        for verb in Verb:
            if hasattr(self, verb.handler_name):
                yield (verb, getattr(self, verb.handler_name))

    def routes(self, tags=None, prefix=None, dependencies=None) -> Iterator[APIRoute]:
        tags = (self._tags or []) + (tags or [])
        prefix = "".join(s for s in [prefix, self._prefix] if s is not None)
        dependencies = (self._dependencies or []) + (dependencies or [])
        for (verb, handler) in self.handlers():
            yield self._route_from_handler(tags, prefix, verb, handler, dependencies)

    def _route_from_handler(self, tags, prefix, verb: Verb, handler, dependencies) -> APIRoute:
        # pylint: disable=too-many-arguments
        kwargs = {
            "path": prefix,
            "methods": [verb.name],
            "endpoint": handler,
            "summary": handler.__doc__,
            "status_code": 200,
            "tags": tags,
            "response_class": JSONResponse,
            "responses": self._RESPONSES,
            "dependencies": dependencies,
        }

        response_model = get_return_type(handler)
        if response_model is not None:
            kwargs["response_model"] = response_model

        return CustomRoute(**kwargs)



import logging
from pydantic import BaseModel
from starlette.requests import Request

from apish.app import Application
from apish.metadata import Contact, Metadata, Version
from apish.routes import Namespace, Resource
from apish.server import run

ns = Namespace([])


log = logging.getLogger(__name__)


class Foo(BaseModel):
    bar: str
    baz: str


LOG_CONFIG = {
    "version": 1,
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "default",
            "level": "INFO",
            "stream": "ext://sys.stdout",
        }
    },
    "root": {"level": "INFO", "handlers": ["console"]},
    "formatters": {
        "default": {
            "format": "%(asctime)s - %(levelname)s - %(name)s - %(message)s",
            "datefmt": "%Y-%m-%d %H:%M:%S",
        }
    },
    "disable_existing_loggers": False,
}


@ns.route("/boom")
class BoomResource(Resource):
    def get(self):
        log.info("Test")
        raise Exception(100 * "boom!")

    async def post(self, request: Request, body: Foo):
        x = await request.body()
        raise Exception(100 * "boom!")


metadata = Metadata(
    title="<title>",
    version=Version(app="v0.1.1", api="v0.1.0"),
    description=None,
    contact=Contact(name="name", url="http://test.com", email=None),
    api_id="49786b4b-1889-46ec-bd72-27f332436e6f",
    audience="company-internal",
)

app = Application("", metadata)
app.add(ns)


run(app, log_config=LOG_CONFIG)



import logging
import subprocess
from functools import lru_cache
from typing import Optional
from pydantic import BaseModel


log = logging.getLogger(__name__)


class Version(BaseModel):
    app: str
    api: str


class Contact(BaseModel):
    name: Optional[str] = None
    url: Optional[str] = None
    email: Optional[str] = None

    def dict(self, *args, **kwargs):
        kwargs.pop("exclude_unset")
        return super().dict(*args, exclude_unset=True, **kwargs)


class Metadata(BaseModel):
    title: str
    version: Version
    description: Optional[str]
    contact: Contact
    api_id: str
    audience: str


@lru_cache(maxsize=1)
def read_app_version():
    """Attempts to read the version from the environment, returning 'NA' when it
    fails.  It first looks for the existence of a VERSION file, then tries to
    use git to obtain a tag.

    :returns: A string describing the version.
    """
    return _read_version_from_file() or _read_version_from_git() or "NA"


def _read_version_from_file():
    handle = None
    try:
        handle = open("VERSION", "r")
        version = handle.read().rstrip()
        log.info("Read version from file")
        return version
    except FileNotFoundError:
        return None
    finally:
        if handle:
            handle.close()


def _read_version_from_git():
    try:
        version = (
            subprocess.run(
                ["git", "describe", "--always", "--tags"], stdout=subprocess.PIPE, check=False
            )
            .stdout.decode("utf-8")
            .rstrip()
        )
        log.info("Read version from git")
        return version
    except subprocess.SubprocessError:
        return None
    except FileNotFoundError:
        return None



import logging

from fastapi import FastAPI, Request
from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse
from fastapi.exception_handlers import RequestValidationError
from starlette.exceptions import HTTPException

from apish.base import Problem
from apish.metadata import Metadata
from apish.routes import Namespace, Resource, RouteGenerator

ns = Namespace(["Generic"])


log = logging.getLogger(__name__)


@ns.route("/metadata")
class Info(Resource):
    async def get(self, request: Request) -> Metadata:
        """Show API metadata"""
        return request.app.metadata


@ns.route("/health")
class Health(Resource):
    async def get(self):
        """Show health status"""
        return {"status": "ok"}


async def from_validation_error(_request: Request, exc: RequestValidationError):
    body = jsonable_encoder(Problem(title="Validation error", status=400, detail=exc.errors()))
    return JSONResponse(status_code=400, content=body)


async def from_http_exception(_request: Request, exc: HTTPException):
    problem = Problem(title=exc.detail, status=exc.status_code, detail={})
    body = jsonable_encoder(problem)
    # pylint: disable=no-member
    return JSONResponse(status_code=problem.status, content=body)


async def from_uncaught_exception(request: Request, exc: Exception):
    problem = Problem(title="Internal Server Error", status=500, detail={})
    body = jsonable_encoder(problem)
    log.error("Unhandled exception (exception: %s, body: %s)", exc, request.state.body)
    # pylint: disable=no-member
    return JSONResponse(status_code=problem.status, content=body)


class Application(FastAPI):
    def __init__(self, root: str, metadata: Metadata, **kwargs):
        self.root = root
        self.metadata = metadata
        exception_handlers = {
            HTTPException: from_http_exception,
            RequestValidationError: from_validation_error,
            Exception: from_uncaught_exception,
        }
        super().__init__(
            root_path="",
            title=metadata.title,
            version=metadata.version.api,
            exception_handlers=exception_handlers,  # type: ignore
            openapi_url=f"{root}/openapi.json",
            docs_url=f"{root}/docs",
            redoc_url=None,
            **kwargs,
        )
        self.add(ns)

    def add(self, route_generator: RouteGenerator) -> None:
        for route in route_generator.routes(prefix=self.root):
            self.routes.append(route)

    def openapi(self):
        openapi = super().openapi()
        if self.metadata.contact:
            openapi["info"]["contact"] = self.metadata.contact.dict(exclude_unset=True)
        if self.metadata.api_id:
            openapi["info"]["x-api-id"] = self.metadata.api_id
        if self.metadata.audience:
            openapi["info"]["x-audience"] = self.metadata.audience
        return openapi



from typing import Optional, Any
from pydantic import BaseModel


class Problem(BaseModel):

    title: str
    status: int
    detail: Any
    type: Optional[str] = None
    instance: Optional[str] = None

    def dict(self, *args, **kwargs):
        kwargs.pop("exclude_unset")
        return super().dict(*args, exclude_unset=True, **kwargs)
